;;;; lastseen.scm - Display when a user was last seen when asked -*- scheme -*-
;;;; Copyright (C) 2002, 2003  Neal H. Walfield  <neal@cs.uml.edu>
;;;;
;;;; This file is part of Deride
;;;;
;;;; Deride is free software; you can redistribute it and/or modify
;;;; it under the terms of the GNU General Public License as published by
;;;; the Free Software Foundation; either version 2, or (at your option)
;;;; any later version.
;;;; 
;;;; Deride is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; GNU General Public License for more details.
;;;; 
;;;; You should have received a copy of the GNU General Public License
;;;; along with the Deride; see the file COPYING.  If not, write to
;;;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

(define-module (backend lastseen))

(use-modules (ice-9 regex))
(use-modules (comm hooks))
(use-modules (comm send))
(use-modules (conf deride))
(use-modules (forum hooks))
(use-modules (forum users))

(use-modules (utils utils))

;; The seen database filename.
(define-public seen-db #f)

;; Frequency with which to flush the database to disk.
;; (in seconds; default of 5 minutes)
(define-public seen-db-flush-frequency (* 5 60))

;; Last time the database was flushed.
(define seen-db-last-flush 0)

;; Flush the seen database to a file.
(define (flush-seen)
  (if (and seen-db
	   (> (- (current-time) seen-db-last-flush)
	      seen-db-flush-frequency))
      (let
	  ((seen-fd (open seen-db (logior O_WRONLY O_CREAT))))
	(if seen-fd
	    (begin
	      (write last-seen-list seen-fd)
	      (set! seen-db-last-flush (current-time))
	      (close seen-fd))))))

;; Map of names to times.  Keys should compare equal using `eqv?'.
(define-public last-seen-list '())

;; Get the data associated with a user.
(define-public (get-last-seen user)
  ;; See if we need to load the seen database.
  (if (and (= seen-db-last-flush 0)
	   seen-db)
      (let
	  ((seen-fd
	    (catch
	     #t
	     (lambda () (open seen-db O_RDONLY))
	     (lambda args
	       (run-hook comm-global-status-message-hook
			 (strerror (car (list-ref args 4))))
	       #f))))
	(if seen-fd
	    (begin
	      (set! last-seen-list (read seen-fd))
	      (close seen-fd)))))

  ;; Do the work.
  (assoc-ref last-seen-list (string-downcase user)))

;; Set the data associated with a user.
(define-public (set-last-seen user data)
  (set! last-seen-list
	(assoc-set! last-seen-list (string-downcase user) data)))

;; Get the user data for a forum
(define-public (get-user-last-seen-in-forum user forum)
  (let ((user-list (get-last-seen user)))
    (if user-list
	(assoc-ref user-list forum)
	#f)))

;; Set one data for a user
(define-public (set-user-last-seen-in-forum user forum data)
  (let* ((forum (string-downcase forum))
	 (user-list (or (get-last-seen user) '()))
	 (list (assoc-set! user-list forum data)))
    (set-last-seen user list)
    (flush-seen)))

;; The format of a seen query.
(define seen-re (make-regexp
		 (string-append
		  "^ *"
		  "(" deride-nick " *[:, ]? *)?"
		  "(Have you|Has anyone|Has anybody)?"
		  " *seen +([a-zA-Z0-9_-]+)")
		 regexp/icase))

(define (tell-last-seen-to-forum forum asker user)
  (let* ((last-saw (get-user-last-seen-in-forum user forum))
	 (since (- (current-time) (or last-saw 0))))
    (cond
     ; Asking about himself.
     ((string-ci=? asker user)
      (send-to-forum forum
		     (string-append
		      asker ": "
		      (random-element-from-list
		       '("Try looking in the mirror."
			 "Never."
			 "Maybe you need to ask yourself that question."
			 "Feeling okay?"
			 "You seem confused.")))))

     ; Asking about me?
     ((string-ci=? deride-nick user)
      (send-to-forum forum
		     (string-append
		      asker ": "
		      (random-element-from-list
		       '("Everyday."
			 "I cannot get enough of that beautiful face!"
			 "Whenever I get a chance."
			 "Right in front of you; and isn't he sexy!")))))

     ; Is in the forum, however, has never spoken.
     ((and (not last-saw)
	   (forum-user-is-present forum user))
      (send-to-forum forum
		     (string-append
		      asker ": "
		     (random-element-from-list
		      (list
		       (string-append user " is quietly lurking.")
		       (string-append user " has been lurking.")
		       (string-append "Apparently, " user
			" is not the talkative type."))))))

     ; Never seen.
     ((not last-saw)
      (send-to-forum forum
		     (string-append
		      asker ": " 
		      (random-element-from-list
		       (list "Nope."
			     "Not recently."
			     (string-append "I have not seen " user
					    " online."))))))

     ; Seen, but has left
     ((not (forum-user-is-present forum user))
      (send-to-forum
       forum
       (string-append asker ": " user
	(cond
	 ((< since 60) " just left.")
	 ((< since (* 15 60))
	  (string-append " left "
			 (number->string (inexact->exact (/ since 60)))
			 " minutes ago."))
	 ((< since (* 2 60 60))
	  (string-append " was here "
			 (number->string (inexact->exact (/ since (* 60 60))))
			 " hours ago."))
	 (else 
	  (string-append " was last on "
			 (number->string (inexact->exact
					  (/ since (* 24 60 60))))
			 " days ago."))))))

     ; Said something less than a minute ago.
     ((< since (* 1 60))
      (send-to-forum forum
		     (string-append
		      asker ": " 
		      (random-element-from-list
		       '("Look around."
			 "Open your eyes."
			 "Umm, right next to you.")))))

     ; Said something less than 10 minutes ago.
     ((< since (* 10 60))
      (send-to-forum forum
		     (string-append
		      asker ": " user
		      (random-element-from-list
		       '(" has been chatting off and on."
			 " was talking just a few minutes ago.")))))

     ; Said something less than an hour ago.
     ((< since (* 60 60))
      (send-to-forum
       forum
       (string-append
	asker ": "  user
	(random-element-from-list
	 (list " took off a while ago."
	       (string-append " was here about "
			      (number->string
			       (* 5 (truncate
				     (inexact->exact (/ since (* 5 60))))))
			      " minutes ago."))))))

     ; In the last day
     ((< since (* 24 60 60))
      (let ((t (truncate (inexact->exact (/ since (* 60 60))))))
	(send-to-forum forum
		       (string-append asker ": I have not seen "
				      user " for "
				      (if (= t 1)
					  "over an hour."
					  (string-append
					   (number->string t)
					   " hours."))))))

     ; Been gone for at least a day.
     (else
      (let ((t (truncate (inexact->exact (/ since
					    (* 24 60 60))))))
	(send-to-forum forum
		       (string-append asker ": " user " has been idle for "
				      (if (= t 1)
					  "over a day."
					  (string-append
					   (number->string t)
					   " days.")))))))))

(define (tell-last-seen-to-user asker user)
  (let ((list (get-last-seen user)))
    (if (not list)
	;; User data is not available.
	(send-to-user asker (string-append user " has never spoken."))
	;; Iterate over the user's data and figure out when the user
	;; said something and where.
	(for-each
	 (lambda (l)
	   (let* ((forum (car l))
		  (last-saw (cdr l))
		  (dt (- (current-time) last-saw))
		  (days (inexact->exact
			 (truncate (/ dt (* 24 60 60)))))
		  (dt (- dt (* days 24 60 60)))
		  (hours (inexact->exact (truncate
					  (/ dt (* 60 60)))))
		  (dt (- dt (* hours (* 60 60))))
		  (minutes (inexact->exact (truncate
					    (/ dt 60)))))
	     (send-to-user
	      asker
	      (string-append user " last spoke on "
			     forum " " 
			     (if (> days 1)
				 (string-append
				  (number->string days)
				  " days, ")
				 "")
			     (if (= days 1)
				 "1 day, "
				 "")
			     (if (> hours 1)
				 (string-append
				  (number->string hours)
				  " hours, ")
				 "")
			     (if (= hours 1)
				 "1 hour, "
				 "")
			     (number->string minutes)
			     (if (= minutes 1)
				 " minute"
				 " minutes")
			     " ago."))))
	 list))))

(add-hook! comm-received-forum-message-hook
	   (lambda (sender forum say msg)
	     (if (eq? (string-ref forum 0) #\#)
		 (begin
		   (let ((match (regexp-exec seen-re msg)))
		     (if match
			 (tell-last-seen-to-forum
			  forum sender (match:substring match 3))))
		   (set-user-last-seen-in-forum sender forum
						(current-time))))))

(define (display-s obj)
  (call-with-output-string
   (lambda (p)
     (display obj p))))

(add-hook! comm-received-personal-message-hook
	   (lambda (sender say msg)
	     (if (string-ci=? msg "seen dump")
		 (send-to-user sender (display-s last-seen-list))
		 (let ((match (regexp-exec seen-re msg)))
		   (if match
		       (tell-last-seen-to-user sender
					       (match:substring match 3)))))))


; (add-hook! comm-entered-forum
;	   (lambda (forum users)

; (add-hook! forum-user-joined-hook
; 	   (lambda (user forum reason)

; (add-hook! forum-user-parted-hook
; 	   (lambda (user forum reason)

; (add-hook! forum-user-quit-hook
; 	   (lambda (user forum-list reason)

; (add-hook! forum-user-schizophrenia-hook
; 	   (lambda (old new forum-list reason) 
