;;;; users.scm - Maintain the members participating in each forum -*- scheme -*-
;;;; Copyright (C) 2001, 2003  Neal H Walfield  <neal@cs.uml.edu>
;;;;
;;;; This file is part of Deride
;;;;
;;;; Deride is free software; you can redistribute it and/or modify
;;;; it under the terms of the GNU General Public License as published by
;;;; the Free Software Foundation; either version 2, or (at your option)
;;;; any later version.
;;;; 
;;;; Deride is distributed in the hope that it will be useful,
;;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;; GNU General Public License for more details.
;;;; 
;;;; You should have received a copy of the GNU General Public License
;;;; along with the Deride; see the file COPYING.  If not, write to
;;;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

(define-module (forum users))

(use-modules (forum hooks))
(use-modules (comm hooks))

;; An alist forum names pointing to a list of strings of users.  Note,
;; forum and user names are case-insenstive.
(define forum-list '())

;; USER has joined FORUM.  If USER is not already on the list, add
;; him, run forum-user-joined-hook and return #t, otherwise, return
;; #f.
(define-public (forum-user-joined forum user reason)
  (let* ((forum (string-downcase forum))
	 (user (string-downcase user))
	 (user-list (assoc forum forum-list)))
    (if user-list
	;; The forum exists.
	(begin
	  (if (not (member user user-list))
	      ;; But the user has not yet been added.
	      (set! forum-list
		    (assoc-set! forum-list forum (cons user (cdr user-list)))))
	  (run-hook forum-user-joined-hook user forum reason)
	  #t)
	;; This is first user in the forum.
	(begin
	  (set! forum-list (acons forum (list user) forum-list))
	  #f))))

(add-hook! comm-entered-forum
	   (lambda (forum users)
	     (letrec
		 ((downcase (lambda (s l)
			      (if (eqv? l '())
				  s
				  (downcase
				   (append s (list (string-downcase (car l))))
				   (cdr l))))))
	       (set! forum-list
		     (assoc-set! forum-list (string-downcase forum)
				 (downcase '() users))))))

;; Internal utility function: remove a user from a forum.
(define (user-left forum user)
  (let* ((forum (string-downcase forum))
	 (user (string-downcase user))
	 (user-list (assoc forum forum-list)))
    (if (and user-list (member user user-list))
	(begin (set! user-list (delete user (cdr user-list)))
	       (assoc-set! forum-list forum user-list)
	       #t)
	#f)))

;; USER has parted FORUM.  If he really was a member of the forum,
;; remove him from the user list, run the forum-user-parted-hook and
;; return #t, otherwise, return #f.
(define-public (forum-user-parted forum user reason)
  (let ((rc (user-left forum user)))
    (if rc
	(run-hook forum-user-parted-hook user forum reason))
    rc))

;; USER has quit, i.e. parted all forums.  Remove him from all the
;; forums and run the forum-user-quit-hook.
(define-public (forum-user-quit user reason)
  (letrec ((loop (lambda (list)
		   (if (eq? list '())
		       '()
		       (if (user-left (caar list) user)
			   (cons (caar list) (loop (cdr list)))
			   (loop (cdr list)))))))
    (run-hook forum-user-quit-hook user (loop forum-list) reason)))

;; USER changed their nick to NEWNAME.  Update all forums and run the
;; forum-user-schizophrenia-hook hook.
(define-public (forum-user-schizophrenia user newname reason)
  (letrec ((loop (lambda (list)
		   (if (eqv? list '())
		       '()
		       (if (member user (cdar list))
			   (begin
			     (let* ((forum (caar list))
				    (user-list (cons newname
						     (delete user
							     (cdar list)))))
			       (assoc-set! forum-list forum user-list)
			       (cons forum (loop (cdr list)))))
			   (loop (cdr list)))))))
    (run-hook forum-user-schizophrenia-hook
	      user newname (loop forum-list) reason)))

;; Return a list of strings containing the users in FORUM.
(define-public (forum-user-list forum)
  (assoc-ref forum-list (string-downcase forum)))

;; Return true if USER is present in FORUM, otherwise true.
(define-public (forum-user-is-present forum user)
  (if (member (string-downcase user)
	      (forum-user-list (string-downcase forum)))
      #t
      #f))